// Code viewer sample plugin

//---------------------------------------------------------------------------
#include <ida.hpp>
#include <idp.hpp>
#include <loader.hpp>
#include <kernwin.hpp>

static const char *sample_code[] =
{
  COLSTR("// some sample code", SCOLOR_CREF),
  "",
  COLSTR("#include", SCOLOR_IMPNAME) COLSTR(" <stdio.h>", SCOLOR_DATNAME),
  "",
  COLSTR("int", SCOLOR_DEFAULT) COLSTR(" main()", SCOLOR_DATNAME),
  COLSTR("{", SCOLOR_DATNAME),
  COLSTR("  printf(", SCOLOR_DATNAME) 
  COLSTR("\"Hello World\\n\"", SCOLOR_SEGNAME) 
  COLSTR(");", SCOLOR_DATNAME),
  "  " COLSTR("return", SCOLOR_DEFAULT) COLSTR(" 0;", SCOLOR_DATNAME),
  COLSTR("}", SCOLOR_DATNAME)
};

// Structure to keep all information about the our sample view
struct sample_info_t
{
  TForm *form;
  TCustomControl *cv;
  TCustomControl *codeview;
  strvec_t sv;
  sample_info_t(TForm *f) : form(f), cv(NULL) {}
};

//--------------------------------------------------------------------------
static bool idaapi lines_linenum(
        TCustomControl * /*cv*/,
        const place_t *p,
        uval_t *num,
        void * /*ud*/)
{
  *num = p->touval(NULL) + 1;
  return true;
}

//--------------------------------------------------------------------------
static int idaapi lines_icon(
        TCustomControl * /*cv*/,
        const place_t *p,
        int *x,
        void * /*ud*/)
{
  bool b = p->touval(NULL) == 6 && *x == 0;
  // setting the highest bit signals that there's another icon 
  // to draw on the current line
  int icon_id = *x == 0 ? 51 : 177;
  return b ? 0x80000000 | icon_id : icon_id;
}

//--------------------------------------------------------------------------
static void idaapi lines_click(
        TCustomControl * /*cv*/,
        const place_t *p,
        int pos,
        int /*shift*/,
        void * /*ud*/)
{
  if ( p != NULL )
    msg("Line click at position: %d\n", pos);
  else
    msg("Click occurred outside of a line\n");
}

//--------------------------------------------------------------------------
int idaapi ui_callback(void *ud, int code, va_list va)
{
  sample_info_t *si = (sample_info_t *)ud;
  switch ( code )
  {
    // how to implement a simple hint callback
    case ui_get_custom_viewer_hint:
      {
        TCustomControl *viewer = va_arg(va, TCustomControl *);
        place_t *place         = va_arg(va, place_t *);
        int *important_lines   = va_arg(va, int *);
        qstring &hint          = *va_arg(va, qstring *);
        if ( si->cv == viewer ) // our viewer
        {
          if ( place == NULL )
            return 0;
          simpleline_place_t *spl = (simpleline_place_t *)place;
          hint.sprnt("Hint for line %ld", spl->n);
          *important_lines = 1;
          return 1;
        }
        break;
      }
    case ui_tform_invisible:
      {
        TForm *f = va_arg(va, TForm *);
        if ( f == si->form )
        {
          delete si;
          unhook_from_notification_point(HT_UI, ui_callback, NULL);
        }
      }
      break;
  }
  return 0;
}

//---------------------------------------------------------------------------
// Create a custom view window
void idaapi run(int)
{
  HWND hwnd = NULL;
  TForm *form = create_tform("Sample code view", &hwnd);
  if ( hwnd == NULL )
  {
    warning("Could not create code view window\n"
            "perhaps it is open?\n"
            "Switching to it.");
    form = find_tform("Sample code view");
    if ( form != NULL )
      switchto_tform(form, true);
    return;
  }
  // allocate block to hold info about our sample view
  sample_info_t *si = new sample_info_t(form);
  // prepare the data to display. we could prepare it on the fly too.
  // but for that we have to use our own custom place_t class decendant.
  for ( int i = 0; i < qnumber(sample_code); i++ )
    si->sv.push_back(simpleline_t(sample_code[i]));
  // create two place_t objects: for the minimal and maximal locations
  simpleline_place_t s1;
  simpleline_place_t s2(si->sv.size()-1);
  // create a custom viewer
  si->cv = create_custom_viewer("", NULL, &s1, &s2, &s1, 0, &si->sv);
  // create a code viewer container for the custom viewer
  si->codeview = create_code_viewer(form, si->cv, CDVF_LINEICONS);
  // set handlers for the code viewer
  set_code_viewer_line_handlers(si->codeview, lines_click, NULL, NULL, lines_icon, lines_linenum);
  // draw maximal 2 icons in the lines widget
  set_code_viewer_lines_icon_margin(si->codeview, 2);
  // also set the ui event callback
  hook_to_notification_point(HT_UI, ui_callback, si);
  // finally display the form on the screen
  open_tform(form, FORM_TAB|FORM_MENU|FORM_RESTORE);
}

//--------------------------------------------------------------------------
int idaapi init(void)
{
  return PLUGIN_KEEP;
}

//--------------------------------------------------------------------------
void idaapi term(void)
{
}

//--------------------------------------------------------------------------
//
//      PLUGIN DESCRIPTION BLOCK
//
//--------------------------------------------------------------------------
plugin_t PLUGIN =
{
  IDP_INTERFACE_VERSION,
  0,                    // plugin flags
  init,                 // initialize

  term,                 // terminate. this pointer may be NULL.

  run,                  // invoke plugin

  "",                   // long comment about the plugin
                        // it could appear in the status line
                        // or as a hint

  "",                    // multiline help about the plugin

  "Sample codeview",    // the preferred short name of the plugin
  ""                    // the preferred hotkey to run the plugin
};
